# storageCore Integration with a BMC Application

This document provides basic steps for integrating storageCore with a BMC application.

Prerequisites:

   For Arm_OpenBMC:
      - storageCore SDK package: sa_storageCore_headers_libs_*.zip
      - storageCore documentation: sa_storageCore_doxygen_*.zip
      - Arm v6 compatible toolchain

   For Arm64_OpenBMC:
      - storageCore SDK package: sa_storageCore_headers_libs_*.zip
      - storageCore documentation: sa_storageCore_doxygen_*.zip
      - ARM 64-bit toolchain

# Steps for Integration

## Extract the contents of the package

Extract the contents of the sa_storageCore_headers_libs_*.zip file and the storageCore_sysCore_template_*.zip file. The following files are necessary for integrating and developing a BMC application:

```
sa_storageCore_headers_libs_*.zip
|-- Arm_OpenBMC
|   `-- storageCore_lib_arm-openbmc.tar.gz
|       `-- arm-openbmc
|           `-- lib
|               |-- libdiagnosticcore.so.*
|               |-- libdiagnosticcore_imp.a
|               |-- libstoragecore.so.*
|               |-- libstoragecore_imp.a
|               `-- libsyscore_common_imp.a
|-- Arm64_OpenBMC
|   `-- storageCore_lib_arm-openbmc-64.tar.gz
|       `-- arm-openbmc-64
|           `-- lib
|               |-- libdiagnosticcore.so.*
|               |-- libdiagnosticcore_imp.a
|               |-- libstoragecore.so.*
|               |-- libstoragecore_imp.a
|               `-- libsyscore_common_imp.a
`-- storageCore_*_headers
    `-- storageCore_headers.tar.gz

storageCore_sysCore_template_*.zip
`-- src
    `-- syscore_template_imp
        |-- sysc_ipc.c
        |-- sysc_synchronization.c
        |-- sysc_template.c
        `-- sysc_threads.c
```

## Implement sysCore template

In `syscore_template.c`, the following functions and macros need to be implemented for basic functionality in a BMC application. Functions that are defined in `sysc_ipc.c`, `sysc_synchronization.c`, and `sysc_threads.c` can be implemented for further functionality.

Note: This is designed to work in synchronous mode.

---

```c
BMC_MAX_TRANSMISSION_BUFFER_SIZE
BMC_MAX_RECEIVE_BUFFER_SIZE
```

---

```c
void _SC_GetControllerDescriptorList(DeviceDescriptorList descriptor_list)
```

Discover and register Smart RAID/HBA controllers.

Create a `device_descriptor` for each adapter and add it to the provided `descriptor_list` by calling `_SC_AddDescriptorToList(DeviceDescriptorList descriptor_list, DeviceDescriptor device_descriptor)`.

Required properties for a `device_descriptor` are:

|Property                   |Description            |
|---------------------------|-----------------------|
|kDeviceInfoBaseConfigNode  |BaseConfig (MCTP EIDs) |
|kDeviceInfoOSDeviceNode    |Device node            |
|kDeviceInfoProtocolType    |Protocol type          |
|kDeviceInfoProtocolAddress |Protocol address       |
|kDeviceInfoDeviceType      |Device Type            |
|kDeviceInfoIndex           |BMIC index             |
|kDeviceInfoPCILocation     |PCI Location           |

See `psysc_descriptor.h` in storageCore_headers for helpful macros in creating `device_descriptors`.

---

```c
SA_DWORD _SC_SendMCTPMessage(
  DeviceDescriptor device_descriptor,
  SA_UINT8 pci_eid,
  SA_UINT8 i2c_eid,
  SA_UINT32 request_length,
  const SA_UINT8 *request,
  SA_UINT32 *response_length,
  SA_UINT8 *response)
)
```

Send requests to PCI/I2C MCTP endpoints. Either PCI VDM or I2C may be used for sending and receiving the packets. If both methods are supported, I2C will be used as a fallback in the case that the PCI VDM method fails.

This function takes the following input parameters:
|Parameter         |Description                              |
|------------------|-----------------------------------------|
|device_descriptor |`device_descriptor` of the target device |
|pci_eid           |PCI endpoint ID of the device            |
|i2c_eid           |I2C endpoint ID of the device            |
|request_length    |Length of the request data               |
|request           |Pointer to the request data              |
|response_length   |Pointer to the length of the response buffer; after sending the request, this will be updated to reflect the length of the response data |
|response          |Pointer to the response buffer           |

This function shall return 0 if the MCTP message was sent successfully, and non-zero otherwise.

---

## Compile sysCore

   For Arm_OpenBMC:

       Using an Arm v6 compatible toolchain, compile sysCore and generate a library (e.g., libsyscore_bmc.a). See [Sample Makefile](#Sample-Makefile) below.

   For Arm64_OpenBMC:

       Using an ARM 64-bit compatible toolchain, compile sysCore and generate a library (e.g., libsyscore_bmc.a). See [Sample Makefile](#Sample-Makefile) below.

## Link storageCore libraries with BMC application

Extract the following libraries from the SDK package:

- libstoragecore_imp.a
- libsyscore_common_imp.a

Link the BMC application with the three libraries (libstoragecore, libsyscore_common, and libsyscore_bmc).

## BMC application development using storageCore

Refer to the storageCore API Doxygen documentation for available functions (extract the documentation from sa_storageCore_doxygen_*.zip). Headers are included in the storageCore SDK package.

## Sample Makefile

The Makefile below assumes the following toolchain and toolchain installation directory:


                     |Toolchain                 |Toolchain Installation Directory |
                     |--------------------------|---------------------------------|
   For Arm_OpenBMC   |arm-openbmc-linux-gnueabi |/opt/crosstool/oecore-x86_64     |
   For Arm64_OpenBMC |aarch64-linux-gnu         |/usr                             |

### Makefile
```
MKDIR              := mkdir
RMDIR              := rm -rf
ifeq ($(PLATFORM),arm-openbmc)
XCOMPILER_PLATFORM ?= i686
XCOMPILER_NAME_ARM ?= arm-openbmc-linux-gnueabi
XCOMPILER_HOME     ?= /opt/crosstool/oecore-x86_64
GPLUSPLUS          ?= $(XCOMPILER_HOME)/sysroots/x86_64-oesdk-linux/usr/bin/$(XCOMPILER_NAME_ARM)/$(XCOMPILER_NAME_ARM)-g++
SYSROOT            := $(XCOMPILER_HOME)/sysroots/arm1176jzs-openbmc-linux-gnueabi
CC                 := $(GCC)
AR                 := $(XCOMPILER_HOME)/sysroots/x86_64-oesdk-linux/usr/bin/$(XCOMPILER_NAME_ARM)/$(XCOMPILER_NAME_ARM)-ar
else ifeq ($(PLATFORM),arm-openbmc-64)
XCOMPILER_PLATFORM ?= i686
XCOMPILER_NAME_ARM ?= aarch64-linux-gnu
XCOMPILER_HOME     ?= /usr
GPLUSPLUS          ?= $(XCOMPILER_HOME)/bin/$(XCOMPILER_NAME_ARM)-gcc
#SYSROOT           := $(XCOMPILER_HOME)/$(XCOMPILER_NAME_ARM)/lib
SYSROOT            := /
endif


# Directory where test application binary will be generated
BIN                := ./bin
# Directory for object files of test application sources
OBJ                := ./obj
# Directory for header files. Copy here all header files from SDK
INCLUDE            := ./inc
# Test application source directory
SRC                := ./src
# sysCore module source directory
SYSCSRC            := ./sysc_src
# Directory for object files of sysCore modules
SYSCOBJ            := ./sysc_obj
# Directory for libraries. Copy here dependent library files from SDK libstoragecore.a, libsyscore_common.a
# sysCore host static library will be generated here
LIBDIR             := ./lib

# Generated test application
EXE                := $(BIN)/bmc_test_application
# Generated sysCore host library
SYSC_LIB           := $(LIBDIR)/libsyscore_bmc.a

CFLAGS             := -I$(INCLUDE)--sysroot=$(SYSROOT) $(ARCHFLAGS)
ARFLAGS            := rcs
LDFLAGS            := -L$(LIBDIR) --sysroot=$(SYSROOT)
ARCHFLAGS          := -march=armv6
LDLIBS             := -lstoragecore_imp -lsyscore_common_imp -lsyscore_bmc

SRCS               := $(wildcard $(SRC)/*.c)
OBJS               := $(patsubst $(SRC)/%.c,$(OBJ)/%.o,$(SRCS))
SYSC_SRCS          := $(wildcard $(SYSCSRC)/*.c)
SYSC_OBJS          := $(patsubst $(SYSCSRC)/%.c,$(SYSCOBJ)/%.o,$(SYSC_SRCS))

all: $(EXE)

$(SYSC_LIB): $(SYSC_OBJS)
             $(AR) $(ARFLAGS) $@ $^

$(SYSCOBJ)/%.o: $(SYSCSRC)/%.c | $(SYSCOBJ)
                $(CC) $(CFLAGS) -c $< -o $@

$(EXE): $(OBJS) | $(BIN) $(SYSC_LIB)
        $(CC) $(LDFLAGS) -o $@ $^ $(LDLIBS)

$(OBJ)/%.o: $(SRC)/%.c | $(OBJ)
            $(CC) $(CFLAGS) -c $< -o $@

$(BIN) $(OBJ) $(SYSCOBJ):
       $(MKDIR) $@
```